@@ -1,11 +1,8 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of this
 # repository contains the full copyright notices and license terms.
 from trytond.pool import Pool
-from . import asset
-from . import product
-from . import invoice
-from . import account
-from . import purchase
+
+from . import account, asset, invoice, product, purchase
 
 
 def register():
@@ -14,7 +11,7 @@ def register():
         asset.AssetLine,
         asset.AssetUpdateMove,
         asset.CreateMovesStart,
-        asset.UpdateAssetStart,
+        asset.AssetRevision,
         asset.UpdateAssetShowDepreciation,
         asset.PrintDepreciationTableStart,
         product.Category,
