@@ -1,23 +1,24 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of this
 # repository contains the full copyright notices and license terms.
-import datetime
 import calendar
+import datetime
 from decimal import Decimal
-from dateutil import relativedelta
-from dateutil import rrule
 from itertools import groupby
 
+from dateutil import relativedelta, rrule
+from sql.functions import CharLength
+
 from trytond.i18n import gettext
-from trytond.model import Workflow, ModelSQL, ModelView, fields, Unique
+from trytond.model import Index, ModelSQL, ModelView, Unique, Workflow, fields
 from trytond.model.exceptions import AccessError
-from trytond.pyson import Eval, Bool, If
-from trytond.pool import Pool
-from trytond.tools import cached_property
-from trytond.transaction import Transaction
-from trytond.wizard import (Wizard, StateView, StateTransition, StateReport,
-    Button)
-from trytond.tools import grouped_slice
 from trytond.modules.company import CompanyReport
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool
+from trytond.pyson import Bool, Eval, If
+from trytond.tools import cached_property, grouped_slice
+from trytond.transaction import Transaction, check_access, without_check_access
+from trytond.wizard import (
+    Button, StateReport, StateTransition, StateView, Wizard)
 
 from .exceptions import PrintDepreciationTableError
 
@@ -53,7 +54,7 @@ class Asset(Workflow, ModelSQL, ModelView):
     'Asset'
     __name__ = 'account.asset'
     _rec_name = 'number'
-    number = fields.Char('Number', readonly=True, select=True)
+    number = fields.Char("Number", readonly=True)
     product = fields.Many2One('product.product', 'Product', required=True,
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
@@ -61,7 +62,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         context={
             'company': Eval('company', None),
             },
-        depends=['state', 'company'],
+        depends={'company'},
         domain=[
             ('type', '=', 'assets'),
             ('depreciable', '=', True),
@@ -81,8 +82,7 @@ class Asset(Workflow, ModelSQL, ModelView):
             ],
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
-            },
-        depends=['product', 'state', 'company'])
+            })
     customer_invoice_line = fields.Function(fields.Many2One(
             'account.invoice.line', 'Customer Invoice Line'),
         'get_customer_invoice_line')
@@ -90,98 +90,81 @@ class Asset(Workflow, ModelSQL, ModelView):
         states={
             'readonly': Eval('state') != 'draft',
             },
-        depends=['state'],
         domain=[('type', '=', 'asset')],
         required=True)
     company = fields.Many2One('company.company', 'Company',
         states={
             'readonly': Eval('state') != 'draft',
             },
-        depends=['state'],
         required=True)
     currency = fields.Function(fields.Many2One('currency.currency',
         'Currency'), 'on_change_with_currency')
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'on_change_with_currency_digits')
-    quantity = fields.Float('Quantity',
-        digits=(16, Eval('unit_digits', 2)),
+    quantity = fields.Float(
+        "Quantity", digits='unit',
         states={
             'readonly': (Bool(Eval('supplier_invoice_line', 1))
                 | Eval('lines', [0])
                 | (Eval('state') != 'draft')),
-            },
-        depends=['state', 'unit_digits'])
+            })
     unit = fields.Many2One('product.uom', 'Unit',
         states={
             'readonly': (Bool(Eval('product'))
                 | (Eval('state') != 'draft')),
-            },
-        depends=['state'])
-    unit_digits = fields.Function(fields.Integer('Unit Digits'),
-        'on_change_with_unit_digits')
-    value = fields.Numeric('Value',
-        digits=(16, Eval('currency_digits', 2)),
+            })
+    value = Monetary(
+        "Value", currency='currency', digits='currency',
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
-        depends=['currency_digits', 'state'],
         required=True,
         help="The value of the asset when purchased.")
-    depreciated_amount = fields.Numeric("Depreciated Amount",
-        digits=(16, Eval('currency_digits', 2)),
+    depreciated_amount = Monetary(
+        "Depreciated Amount", currency='currency', digits='currency',
         domain=[
             ('depreciated_amount', '<=', Eval('value')),
             ],
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
-        depends=['currency_digits', 'value', 'state'],
         required=True,
         help="The amount already depreciated at the start date.")
-    depreciating_value = fields.Function(fields.Numeric(
-            "Depreciating Value",
-            digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits'],
+    depreciating_value = fields.Function(Monetary(
+            "Depreciating Value", currency='currency', digits='currency',
             help="The value of the asset at the start date."),
         'on_change_with_depreciating_value')
-    residual_value = fields.Numeric('Residual Value',
+    residual_value = Monetary(
+        "Residual Value", currency='currency', digits='currency',
+        required=True,
         domain=[
             ('residual_value', '<=', Eval('depreciating_value')),
             ],
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
-            },
-        depends=['currency_digits', 'depreciating_value', 'state'],
-        required=True,
-        digits=(16, Eval('currency_digits', 2)))
+            })
     purchase_date = fields.Date('Purchase Date', states={
             'readonly': (Bool(Eval('supplier_invoice_line', 1))
                 | Eval('lines', [0])
                 | (Eval('state') != 'draft')),
             },
-        required=True,
-        depends=['state'])
+        required=True)
     start_date = fields.Date('Start Date', states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
         required=True,
-        domain=[('start_date', '<=', Eval('end_date', None))],
-        depends=['state', 'end_date'])
+        domain=[('start_date', '<=', Eval('end_date', None))])
     end_date = fields.Date('End Date',
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
         required=True,
-        domain=[('end_date', '>=', Eval('start_date', None))],
-        depends=['state', 'start_date'])
+        domain=[('end_date', '>=', Eval('start_date', None))])
     depreciation_method = fields.Selection([
             ('linear', 'Linear'),
             ], 'Depreciation Method',
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
-        required=True,
-        depends=['state'])
+        required=True)
     frequency = fields.Selection([
             ('monthly', 'Monthly'),
             ('yearly', 'Yearly'),
@@ -189,20 +172,18 @@ class Asset(Workflow, ModelSQL, ModelView):
         required=True,
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
-            },
-        depends=['state'])
+            })
     state = fields.Selection([
             ('draft', 'Draft'),
             ('running', 'Running'),
             ('closed', 'Closed'),
-            ], 'State', readonly=True)
+            ], "State", readonly=True, sort=False)
     lines = fields.One2Many('account.asset.line', 'asset', 'Lines',
         readonly=True)
     move = fields.Many2One('account.move', 'Account Move', readonly=True,
         domain=[
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=['company'])
+            ])
     update_moves = fields.Many2Many('account.asset-update-account.move',
         'asset', 'move', 'Update Moves', readonly=True,
         domain=[
@@ -210,23 +191,36 @@ class Asset(Workflow, ModelSQL, ModelView):
             ],
         states={
             'invisible': ~Eval('update_moves'),
-            },
-        depends=['company'])
+            })
     comment = fields.Text('Comment')
+    revisions = fields.One2Many(
+        'account.asset.revision', 'asset', "Revisions", readonly=True)
 
     @classmethod
     def __setup__(cls):
+        cls.number.search_unaccented = False
         super(Asset, cls).__setup__()
         table = cls.__table__()
         cls._sql_constraints = [
             ('invoice_line_uniq', Unique(table, table.supplier_invoice_line),
                 'account_asset.msg_asset_invoice_line_unique'),
             ]
+        cls._sql_indexes.add(
+            Index(
+                table,
+                (table.state, Index.Equality()),
+                where=table.state.in_(['draft', 'running'])))
         cls._transitions |= set((
                 ('draft', 'running'),
                 ('running', 'closed'),
+                ('running', 'draft'),
                 ))
         cls._buttons.update({
+                'draft': {
+                    'invisible': (Eval('lines', [])
+                        | (Eval('state') != 'running')),
+                    'depends': ['state'],
+                    },
                 'run': {
                     'invisible': Eval('state') != 'draft',
                     'depends': ['state'],
@@ -236,13 +230,12 @@ class Asset(Workflow, ModelSQL, ModelView):
                     'depends': ['state'],
                     },
                 'create_lines': {
-                    'invisible': (Eval('lines', [])
-                        | (Eval('state') != 'draft')),
+                    'invisible': ~Eval('state').in_(['draft', 'running']),
                     'depends': ['state'],
                     },
                 'clear_lines': {
                     'invisible': (~Eval('lines', [0])
-                        | (Eval('state') != 'draft')),
+                        | ~Eval('state').in_(['draft', 'running'])),
                     'depends': ['state'],
                     },
                 'update': {
@@ -252,13 +245,9 @@ class Asset(Workflow, ModelSQL, ModelView):
                 })
 
     @classmethod
-    def __register__(cls, module_name):
-        table_h = cls.__table_handler__(module_name)
-
-        # Migration from 3.8: rename reference into number
-        if table_h.column_exist('reference'):
-            table_h.column_rename('reference', 'number')
-        super(Asset, cls).__register__(module_name)
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
 
     @staticmethod
     def default_state():
@@ -314,14 +303,7 @@ class Asset(Workflow, ModelSQL, ModelView):
 
     @fields.depends('company')
     def on_change_with_currency(self, name=None):
-        if self.company:
-            return self.company.currency.id
-
-    @fields.depends('company')
-    def on_change_with_currency_digits(self, name=None):
-        if self.company:
-            return self.company.currency.digits
-        return 2
+        return self.company.currency if self.company else None
 
     @fields.depends('supplier_invoice_line', 'unit')
     def on_change_supplier_invoice_line(self):
@@ -361,15 +343,7 @@ class Asset(Workflow, ModelSQL, ModelView):
 
     @fields.depends('product')
     def on_change_with_unit(self):
-        if not self.product:
-            return None
-        return self.product.default_uom.id
-
-    @fields.depends('unit')
-    def on_change_with_unit_digits(self, name=None):
-        if not self.unit:
-            return 2
-        return self.unit.digits
+        return self.product.default_uom if self.product else None
 
     @fields.depends('end_date', 'product', 'start_date')
     def on_change_with_end_date(self):
@@ -410,7 +384,10 @@ class Asset(Workflow, ModelSQL, ModelView):
         # dateutil >= 2.0 has replace __nonzero__ by __bool__ which doesn't
         # work in Python < 3
         if delta == relativedelta.relativedelta():
-            return [self.end_date]
+            if not self.lines:
+                return [self.end_date]
+            else:
+                return []
         if self.frequency == 'monthly':
             rule = rrule.rrule(rrule.MONTHLY, dtstart=self.start_date,
                 bymonthday=int(config.get_multivalue(
@@ -513,7 +490,6 @@ class Asset(Workflow, ModelSQL, ModelView):
     def create_lines(cls, assets):
         pool = Pool()
         Line = pool.get('account.asset.line')
-        cls.clear_lines(assets)
 
         lines = []
         for asset in assets:
@@ -549,7 +525,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         Move = pool.get('account.move')
         MoveLine = pool.get('account.move.line')
 
-        period_id = Period.find(self.company.id, line.date)
+        period = Period.find(self.company, line.date)
         with Transaction().set_context(date=line.date):
             expense_line = MoveLine(
                 credit=0,
@@ -565,7 +541,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         return Move(
             company=self.company,
             origin=line,
-            period=period_id,
+            period=period,
             journal=self.account_journal,
             date=line.date,
             lines=[expense_line, depreciation_line],
@@ -580,6 +556,8 @@ class Asset(Workflow, ModelSQL, ModelView):
         Move = pool.get('account.move')
         Line = pool.get('account.asset.line')
 
+        cls.create_lines(assets)
+
         moves = []
         lines = []
         for asset_ids in grouped_slice(assets):
@@ -607,8 +585,9 @@ class Asset(Workflow, ModelSQL, ModelView):
         MoveLine = pool.get('account.move.line')
 
         if date is None:
-            date = Date.today()
-        period_id = Period.find(self.company.id, date)
+            with Transaction().set_context(company=self.company.id):
+                date = Date.today()
+        period = Period.find(self.company, date)
         if self.supplier_invoice_line:
             account_asset = self.supplier_invoice_line.account.current()
         else:
@@ -638,7 +617,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         return Move(
             company=self.company,
             origin=self,
-            period=period_id,
+            period=period,
             journal=self.account_journal,
             date=date,
             lines=lines,
@@ -660,6 +639,16 @@ class Asset(Workflow, ModelSQL, ModelView):
                 'asset_sequence', company=asset.company.id).get()
         cls.save(assets)
 
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('draft')
+    def draft(cls, assets):
+        for asset in assets:
+            if asset.lines:
+                raise AccessError(
+                    gettext('account_asset.msg_asset_draft_lines',
+                        asset=asset.rec_name))
+
     @classmethod
     @ModelView.button
     @Workflow.transition('running')
@@ -716,6 +705,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         default.setdefault('number', None)
         default.setdefault('supplier_invoice_line', None)
         default.setdefault('move')
+        default.setdefault('revisions', [])
         return super(Asset, cls).copy(assets, default=default)
 
     @classmethod
@@ -723,7 +713,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         for asset in assets:
             if asset.state != 'draft':
                 raise AccessError(
-                    gettext('account_asset.msg_delete_draft',
+                    gettext('account_asset.msg_asset_delete_draft',
                         asset=asset.rec_name))
         return super(Asset, cls).delete(assets)
 
@@ -734,23 +724,23 @@ class AssetLine(ModelSQL, ModelView):
     asset = fields.Many2One('account.asset', 'Asset', required=True,
         ondelete='CASCADE', readonly=True)
     date = fields.Date('Date', readonly=True)
-    depreciation = fields.Numeric('Depreciation',
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'],
+    depreciation = Monetary(
+        "Depreciation", currency='currency', digits='currency',
         required=True, readonly=True)
-    acquired_value = fields.Numeric('Acquired Value', readonly=True,
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
-    depreciable_basis = fields.Numeric('Depreciable Basis', readonly=True,
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
-    actual_value = fields.Numeric('Actual Value', readonly=True,
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
-    accumulated_depreciation = fields.Numeric(
-        'Accumulated Depreciation', readonly=True,
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
+    acquired_value = Monetary(
+        "Acquired Value", currency='currency', digits='currency',
+        readonly=True)
+    depreciable_basis = Monetary(
+        "Depreciable Basis", currency='currency', digits='currency',
+        readonly=True)
+    actual_value = Monetary(
+        "Actual Value", currency='currency', digits='currency', readonly=True)
+    accumulated_depreciation = Monetary(
+        "Accumulated Depreciation", currency='currency', digits='currency',
+        readonly=True)
     move = fields.Many2One('account.move', 'Account Move', readonly=True)
     currency = fields.Function(fields.Many2One('currency.currency',
         'Currency'), 'on_change_with_currency')
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'on_change_with_currency_digits')
 
     @classmethod
     def __setup__(cls):
@@ -760,20 +750,14 @@ class AssetLine(ModelSQL, ModelView):
 
     @fields.depends('asset', '_parent_asset.currency')
     def on_change_with_currency(self, name=None):
-        if self.asset:
-            return self.asset.currency.id
-
-    @fields.depends('asset', '_parent_asset.currency_digits')
-    def on_change_with_currency_digits(self, name=None):
-        if self.asset:
-            return self.asset.currency_digits
+        return self.asset.currency if self.asset else None
 
 
 class AssetUpdateMove(ModelSQL):
     'Asset - Update - Move'
     __name__ = 'account.asset-update-account.move'
-    asset = fields.Many2One('account.asset', 'Asset', ondelete='CASCADE',
-        select=True, required=True)
+    asset = fields.Many2One(
+        'account.asset', "Asset", ondelete='CASCADE', required=True)
     move = fields.Many2One('account.move', 'Move', required=True)
 
 
@@ -798,9 +782,11 @@ class CreateMoves(Wizard):
             ])
     create_moves = StateTransition()
 
+    @without_check_access
     def transition_create_moves(self):
-        Asset = Pool().get('account.asset')
-        with Transaction().set_context(_check_access=True):
+        pool = Pool()
+        with check_access():
+            Asset = pool.get('account.asset')
             assets = Asset.search([
                     ('state', '=', 'running'),
                     ])
@@ -809,14 +795,6 @@ class CreateMoves(Wizard):
         return 'end'
 
 
-class UpdateAssetStart(ModelView):
-    'Update Asset Start'
-    __name__ = 'account.asset.update.start'
-    value = fields.Numeric('Asset Value', required=True)
-    residual_value = fields.Numeric('Residual Value', required=True)
-    end_date = fields.Date('End Date', required=True)
-
-
 class UpdateAssetShowDepreciation(ModelView):
     'Update Asset Show Depreciation'
     __name__ = 'account.asset.update.show_depreciation'
@@ -826,7 +804,6 @@ class UpdateAssetShowDepreciation(ModelView):
             ('date', '>=', Eval('latest_move_date')),
             ('date', '<=', Eval('next_depreciation_date')),
             ],
-        depends=['latest_move_date', 'next_depreciation_date'],
         help=('The date must be between the last update/depreciation date '
             'and the next depreciation date.'))
     latest_move_date = fields.Date('Latest Move Date', readonly=True)
@@ -841,8 +818,8 @@ class UpdateAssetShowDepreciation(ModelView):
 class UpdateAsset(Wizard):
     'Update Asset'
     __name__ = 'account.asset.update'
-    start = StateView('account.asset.update.start',
-        'account_asset.asset_update_start_view_form', [
+    start = StateView('account.asset.revision',
+        'account_asset.asset_revision_view_form', [
             Button('Cancel', 'end', 'tryton-cancel'),
             Button('OK', 'update_asset', 'tryton-ok', True),
             ])
@@ -860,6 +837,7 @@ class UpdateAsset(Wizard):
             'value': self.record.value,
             'residual_value': self.record.residual_value,
             'end_date': self.record.end_date,
+            'asset': self.record.id,
             }
 
     def transition_update_asset(self):
@@ -898,14 +876,16 @@ class UpdateAsset(Wizard):
         pool = Pool()
         Period = pool.get('account.period')
         Move = pool.get('account.move')
-        period_id = Period.find(asset.company.id, self.show_move.date)
-        return Move(
+        period = Period.find(asset.company, self.show_move.date)
+        move = Move(
             company=asset.company,
             origin=asset,
             journal=asset.account_journal.id,
-            period=period_id,
+            period=period,
             date=self.show_move.date,
             )
+        move.lines = self.get_move_lines(asset)
+        return move
 
     def get_move_lines(self, asset):
         MoveLine = Pool().get('account.move.line')
@@ -930,7 +910,6 @@ class UpdateAsset(Wizard):
         if not (latest_move_date <= self.show_move.date <= next_date):
             raise ValueError('The update move date is invalid')
         move = self.get_move(self.record)
-        move.lines = self.get_move_lines(self.record)
         move.save()
         self.model.write([self.record], {
                 'update_moves': [('add', [move.id])],
@@ -944,10 +923,61 @@ class UpdateAsset(Wizard):
                 'residual_value': self.start.residual_value,
                 'end_date': self.start.end_date,
                 })
+        self.model.clear_lines([self.record])
         self.model.create_lines([self.record])
+        self.start.asset = self.record
+        self.start.save()
         return 'end'
 
 
+class AssetRevision(ModelSQL, ModelView):
+    "Asset Revision"
+    __name__ = 'account.asset.revision'
+    currency = fields.Function(
+        fields.Many2One('currency.currency', "Currency"),
+        'on_change_with_currency')
+    value = Monetary(
+        "Asset Value", currency='currency', digits='currency',
+        required=True)
+    residual_value = Monetary(
+        "Residual Value", currency='currency', digits='currency',
+        required=True)
+    end_date = fields.Date("End Date", required=True)
+    origin = fields.Reference("Origin", selection='get_origins')
+    description = fields.Char("Description")
+    asset = fields.Many2One('account.asset', "Asset", required=True)
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('asset')
+
+    @fields.depends('asset', '_parent_asset.currency')
+    def on_change_with_currency(self, name=None):
+        return self.asset.currency if self.asset else None
+
+    @fields.depends('origin', 'value', 'asset', '_parent_asset.value')
+    def on_change_origin(self, name=None):
+        pool = Pool()
+        InvoiceLine = pool.get('account.invoice.line')
+        if isinstance(self.origin, InvoiceLine) and self.origin.id >= 0:
+            self.value = self.asset.value + self.origin.amount
+
+    @staticmethod
+    def _get_origin():
+        "Return list of Model names for origin Reference"
+        return ['account.invoice.line']
+
+    @classmethod
+    def get_origins(cls):
+        pool = Pool()
+        IrModel = pool.get('ir.model')
+
+        get_name = IrModel.get_name
+        models = cls._get_origin()
+        return [(None, '')] + [(m, get_name(m)) for m in models]
+
+
 class AssetDepreciationTable(CompanyReport):
     'Asset Depreciation Table'
     __name__ = 'account.asset.depreciation_table'
@@ -1117,11 +1147,9 @@ class PrintDepreciationTableStart(ModelView):
     __name__ = 'account.asset.print_depreciation_table.start'
 
     start_date = fields.Date('Start Date', required=True,
-        domain=[('start_date', '<', Eval('end_date'))],
-        depends=['end_date'])
+        domain=[('start_date', '<', Eval('end_date'))])
     end_date = fields.Date('End Date', required=True,
-        domain=[('end_date', '>', Eval('start_date'))],
-        depends=['start_date'])
+        domain=[('end_date', '>', Eval('start_date'))])
 
     @staticmethod
     def default_start_date():
@@ -1142,21 +1170,21 @@ class PrintDepreciationTable(Wizard):
             ])
     print_ = StateReport('account.asset.depreciation_table')
 
+    @check_access
     def do_print_(self, action):
         pool = Pool()
         Asset = pool.get('account.asset')
-        with Transaction().set_context(_check_access=True):
-            assets = Asset.search([
-                    ['OR',
-                        ('purchase_date', '<=', self.start.end_date),
-                        ('start_date', '<=', self.start.end_date),
-                        ],
-                    ['OR',
-                        ('move', '=', None),
-                        ('move.date', '>=', self.start.start_date),
-                        ],
-                    ('state', '!=', 'draft'),
-                    ])
+        assets = Asset.search([
+                ['OR',
+                    ('purchase_date', '<=', self.start.end_date),
+                    ('start_date', '<=', self.start.end_date),
+                    ],
+                ['OR',
+                    ('move', '=', None),
+                    ('move.date', '>=', self.start.start_date),
+                    ],
+                ('state', '!=', 'draft'),
+                ])
         if not assets:
             raise PrintDepreciationTableError(
                 gettext('account_asset.msg_no_assets'))
