@@ -1,14 +1,12 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of this
 # repository contains the full copyright notices and license terms.
-from trytond.model import fields, Unique
-from trytond.pyson import Eval
+from trytond.model import Unique, fields
 from trytond.pool import Pool, PoolMeta
-
-__all__ = ['InvoiceLine']
+from trytond.pyson import Eval
+from trytond.transaction import Transaction
 
 
-class InvoiceLine:
-    __metaclass__ = PoolMeta
+class InvoiceLine(metaclass=PoolMeta):
     __name__ = 'account.invoice.line'
     asset = fields.Many2One('account.asset', 'Asset', domain=[
             ('state', '=', 'running'),
@@ -16,10 +14,9 @@ class InvoiceLine:
             ],
         states={
             'invisible': (~Eval('is_assets_depreciable', False)
-                | (Eval('_parent_invoice', {}).get('type',
-                        Eval('invoice_type')) != 'out')),
-            },
-        depends=['product', 'is_assets_depreciable'])
+                | (Eval('invoice_type') != 'out')),
+            'readonly': Eval('invoice_state') != 'draft',
+            })
     is_assets_depreciable = fields.Function(fields.Boolean(
             'Is Assets depreciable'),
         'on_change_with_is_assets_depreciable')
@@ -30,10 +27,18 @@ class InvoiceLine:
         table = cls.__table__()
         cls._sql_constraints += [
             ('asset_uniq', Unique(table, table.asset),
-                'Asset can be used only once on invoice line!'),
+                'account_asset.msg_invoice_line_asset_unique'),
             ]
 
-    @fields.depends('product', 'invoice', 'invoice_type')
+    @classmethod
+    def _account_domain(cls, type_):
+        domain = super()._account_domain(type_)
+        if type_ == 'in':
+            domain.append(('type.fixed_asset', '=', True))
+        return domain
+
+    @fields.depends('product', 'invoice', 'invoice_type',
+        '_parent_invoice.invoice_date', '_parent_invoice.accounting_date')
     def on_change_product(self):
         super(InvoiceLine, self).on_change_product()
         if self.invoice and self.invoice.type:
@@ -44,7 +49,10 @@ class InvoiceLine:
         if (self.product and type_ == 'in'
                 and self.product.type == 'assets'
                 and self.product.depreciable):
-            self.account = self.product.account_asset_used
+            date = (self.invoice.accounting_date or self.invoice.invoice_date
+                if self.invoice else None)
+            with Transaction().set_context(date=date):
+                self.account = self.product.account_asset_used
 
     @fields.depends('asset', 'unit')
     def on_change_asset(self):
@@ -69,5 +77,6 @@ class InvoiceLine:
     def get_move_lines(self):
         Asset = Pool().get('account.asset')
         if self.asset:
-            Asset.close([self.asset], account=self.account)
+            date = self.invoice.accounting_date or self.invoice.invoice_date
+            Asset.close([self.asset], account=self.account, date=date)
         return super(InvoiceLine, self).get_move_lines()
